
/* Chrysalide - Outil d'analyse de fichiers binaires
 * xor.c - transormation via opération XOR
 *
 * Copyright (C) 2023 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "xor.h"


#include <assert.h>
#include <malloc.h>
#include <stdint.h>
#include <string.h>


#include "../modifier-int.h"



/* ----------------------- RECHERCHE D'UN MOTIF DE TEXTE BRUT ----------------------- */


/* Initialise la classe des transmissions via opération XOR. */
static void g_scan_xor_modifier_class_init(GScanXorModifierClass *klass);

/* Initialise une instance de transmission via opération XOR. */
static void g_scan_xor_modifier_init(GScanXorModifier *);

/* Supprime toutes les références externes. */
static void g_scan_xor_modifier_dispose(GScanXorModifier *);

/* Procède à la libération totale de la mémoire. */
static void g_scan_xor_modifier_finalize(GScanXorModifier *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Fournit le nom d'appel d'un modificateur pour motif. */
static char *g_scan_xor_modifier_get_name(const GScanXorModifier *);

/* Transforme une séquence d'octets pour motif de recherche. */
static bool g_scan_xor_modifier_transform(const GScanXorModifier *, const sized_binary_t *, size_t, sized_binary_t **, size_t *);

/* Détermine si un argument est bien toléré par un modificateur. */
static bool g_scan_xor_modifier_can_handle_arg(const GScanXorModifier *, const modifier_arg_t *);

/* Transforme une séquence d'octets pour motif de recherche. */
static bool g_scan_xor_modifier_transform_with_arg(const GScanXorModifier *, const sized_binary_t *, size_t, const modifier_arg_t *, sized_binary_t **, size_t *);



/* ---------------------------------------------------------------------------------- */
/*                         RECHERCHE D'UN MOTIF DE TEXTE BRUT                         */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une transormation via une opération XOR. */
G_DEFINE_TYPE(GScanXorModifier, g_scan_xor_modifier, G_TYPE_SCAN_TOKEN_MODIFIER);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des transmissions via opération XOR.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_xor_modifier_class_init(GScanXorModifierClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanTokenModifierClass *modifier;      /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_scan_xor_modifier_dispose;
    object->finalize = (GObjectFinalizeFunc)g_scan_xor_modifier_finalize;

    modifier = G_SCAN_TOKEN_MODIFIER_CLASS(klass);

    modifier->get_name = (get_scan_modifier_name_fc)g_scan_xor_modifier_get_name;

    modifier->transform = (transform_scan_token_fc)g_scan_xor_modifier_transform;
    modifier->can_handle = (can_token_modifier_handle_arg)g_scan_xor_modifier_can_handle_arg;
    modifier->transform_with = (transform_scan_token_with_fc)g_scan_xor_modifier_transform_with_arg;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = instance à initialiser.                           *
*                                                                             *
*  Description : Initialise une instance de transmission via opération XOR.   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_xor_modifier_init(GScanXorModifier *modifier)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = instance d'objet GLib à traiter.                  *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_xor_modifier_dispose(GScanXorModifier *modifier)
{
    G_OBJECT_CLASS(g_scan_xor_modifier_parent_class)->dispose(G_OBJECT(modifier));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = instance d'objet GLib à traiter.                  *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_xor_modifier_finalize(GScanXorModifier *modifier)
{
    G_OBJECT_CLASS(g_scan_xor_modifier_parent_class)->finalize(G_OBJECT(modifier));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Construit un modificateur livrant des octets traités par XOR.*
*                                                                             *
*  Retour      : Mécanisme mis en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanTokenModifier *g_scan_xor_modifier_new(void)
{
    GScanTokenModifier *result;                    /* Structure à retourner       */

    result = g_object_new(G_TYPE_SCAN_XOR_MODIFIER, NULL);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à consulter.                         *
*                                                                             *
*  Description : Fournit le nom d'appel d'un modificateur pour motif.         *
*                                                                             *
*  Retour      : Désignation humaine.                                         *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static char *g_scan_xor_modifier_get_name(const GScanXorModifier *modifier)
{
    char *result;                           /* Désignation à retourner     */

    result = strdup("xor");

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à solliciter.                        *
*                src      = séquence d'octets à traiter.                      *
*                scount   = quantité de ces séquences.                        *
*                dest     = nouvelle(s) séquence(s) d'octets obtenue(s) [OUT] *
*                dcount   = quantité de ces séquences.                        *
*                                                                             *
*  Description : Transforme une séquence d'octets pour motif de recherche.    *
*                                                                             *
*  Retour      : Bilan de l'opération : succès ou échec.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_xor_modifier_transform(const GScanXorModifier *modifier, const sized_binary_t *src, size_t scount, sized_binary_t **dest, size_t *dcount)
{
    bool result;                            /* Bilan d'opération à renvoyer*/
    sized_binary_t *binary;                 /* Raccourci vers le stockage  */
    size_t i;                               /* Boucle de parcours #1       */
    const sized_binary_t *_src;             /* Source courante             */
    long long x;                            /* Boucle de parcours #2       */
    size_t k;                               /* Boucle de parcours #3       */

    result = true;

    *dcount = scount * 256;
    *dest = malloc(*dcount * sizeof(sized_binary_t));

    binary = &(*dest)[0];

    for (i = 0; i < scount; i++)
    {
        _src = src + i;

        for (x = 0; x <= 0xff; x++, binary++)
        {
            binary->len = _src->len;
            binary->data = malloc(binary->len);

            for (k = 0; k < _src->len; k++)
                binary->data[k] = _src->data[k] ^ x;

        }

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à solliciter.                        *
*                arg      = argument de personnalisation.                     *
*                                                                             *
*  Description : Détermine si un argument est bien toléré par un modificateur.*
*                                                                             *
*  Retour      : Bilan de la consultation : support ou non.                   *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_xor_modifier_can_handle_arg(const GScanXorModifier *modifier, const modifier_arg_t *arg)
{
    bool result;                            /* Bilan d'opération à renvoyer*/

    switch (arg->type)
    {
        case MAT_UNSIGNED_INTEGER:
            result = (arg->value.u_integer <= UINT8_MAX);
            break;

        case MAT_RANGE:
            result = (INT8_MIN <= arg->value.range.start && arg->value.range.end <= INT8_MAX);
            break;

        default:
            result = false;

    }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : modifier = modificateur à solliciter.                        *
*                src      = séquence d'octets à traiter.                      *
*                scount   = quantité de ces séquences.                        *
*                arg      = argument de personnalisation.                     *
*                dest     = nouvelle(s) séquence(s) d'octets obtenue(s) [OUT] *
*                dcount   = quantité de ces séquences.                        *
*                                                                             *
*  Description : Transforme une séquence d'octets pour motif de recherche.    *
*                                                                             *
*  Retour      : Bilan de l'opération : succès ou échec.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_xor_modifier_transform_with_arg(const GScanXorModifier *modifier, const sized_binary_t *src, size_t scount, const modifier_arg_t *arg, sized_binary_t **dest, size_t *dcount)
{
    bool result;                            /* Bilan d'opération à renvoyer*/
    sized_binary_t *binary;                 /* Raccourci vers le stockage  */
    size_t i;                               /* Boucle de parcours #1       */
    const sized_binary_t *_src;             /* Source courante             */
    long long x;                            /* Boucle de parcours #2       */
    size_t k;                               /* Boucle de parcours #3       */

    result = true;

    switch (arg->type)
    {
        case MAT_UNSIGNED_INTEGER:

            *dcount = scount;
            *dest = malloc(*dcount * sizeof(sized_binary_t));

            binary = &(*dest)[0];

            for (i = 0; i < scount; i++)
            {
                _src = src + i;

                binary->len = _src->len;
                binary->data = malloc(binary->len);

                for (k = 0; k < _src->len; k++)
                    binary->data[k] = _src->data[k] ^ arg->value.u_integer;

            }

            break;

        case MAT_RANGE:

            *dcount = scount * (arg->value.range.end - arg->value.range.start + 1);
            *dest = malloc(*dcount * sizeof(sized_binary_t));

            binary = &(*dest)[0];

            for (i = 0; i < scount; i++)
            {
                _src = src + i;

                for (x = arg->value.range.start; x <= arg->value.range.end; x++, binary++)
                {
                    binary->len = _src->len;
                    binary->data = malloc(binary->len);

                    for (k = 0; k < _src->len; k++)
                        binary->data[k] = _src->data[k] ^ x;

                }

            }

            break;

        default:
            assert(false);
            result = false;

    }

    return result;

}
